# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['remerge']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.18,<1.23', 'tqdm>=4.64.1,<5.0.0']

setup_kwargs = {
    'name': 'remerge-mwe',
    'version': '0.2.0',
    'description': 'REMERGE is a Multi-Word Expression (MWE) discovery algorithm derived from the MERGE algorithm.',
    'long_description': '# REMERGE - Multi-Word Expression discovery algorithm\n\nREMERGE is a Multi-Word Expression (MWE) discovery algorithm, which started as a re-implementation and simplification of a similar algorithm called MERGE, detailed in a publication and PhD thesis[^2][^3]. The primary benefit of this algorithm is that it\'s non-parametric in regards to the size of the n-grams that constitute a MWE—you do not need to specify a priori how many n-grams comprise a MWE—you only need to specify the number of iterations you want the algorithm to run.\n\nThe code was originally derived from an existing implementation from the original author[^1] that I reviewed, converted from python 2 to 3, then modified and updated with the following:\n- a correction of the log-likelihood calculation; previously it was not using the correct values for the contingency table\n- the removal of gapsize / discontinuous bigrams (see below for issues with the prior implementation)\n- an overall reduction in codebase size and complexity\n  - ~60% reduction in loc\n  - removed `pandas` and `nltk` dependencies\n- type annotations\n- the inclusion of additional metrics (Frequency, [NPMI](https://svn.spraakdata.gu.se/repos/gerlof/pub/www/Docs/npmi-pfd.pdf)[^4]) for selecting the winning bigram.\n- corrections for merging sequential bigrams greedily and completely.\n  - e.g. `\'ya ya ya ya\'` -> `\'(ya ya) (ya ya)\'` -> `\'(ya ya ya ya)\'`. Previously the merge order was non-deterministic, and you could end up with `\'ya (ya ya) ya\'`\n- An overall simplification of the algorithm. \n  - As a tradeoff, this version may be less efficient. After a bigram is merged into a single lexeme in the original implementation, new bigrams and conflicting (old) bigrams were respectively added and subtracted from a mutable counter of bigrams. The counts of this object were difficult to track and validate, and resulted in errors in certain cases, so I removed this step. Instead, only the lexeme data is updated with the new merged lexemes, and bigram data is recreated from scratch using the lexeme data. There is certainly improvement for optimization here, particularly in the case where we are recalculating bigrams from lines that did not contain a merged lexeme and thus their bigram data did not change.\n\n#### Usage\n\n```python\nimport remerge\n\ncorpus = [\n    ["a", "list", "of", "already", "tokenized", "texts"],\n    ["where", "each", "item", "is", "a", "list", "of", "tokens"],\n    ["isn\'t", "a", "list", "nice"]\n]\n\nwinners = remerge.run(\n    corpus, iterations=1, method=remerge.SelectionMethod.frequency, progress_bar="all"\n)\n# winners[0].merged_lexeme.word == (\'a\', \'list\')\n```\n\nThere are 3 bigram winner selection methods: [Log-Likelihood (G²)](https://aclanthology.org/J93-1003.pdf)[^5], [NPMI](https://svn.spraakdata.gu.se/repos/gerlof/pub/www/Docs/npmi-pfd.pdf)[^4], and raw frequency. They are available under the `SelectionMethod` enum. The default is log-likelihood, which was used in the original implementation.\n\nIf using `NPMI` (`SelectionMethod.npmi`), you likely want to provide a `min_count` parameter, "as infrequent word pairs tend to dominate the top of bigramme lists that are ranked after PMI". (p. 4[^4])\n\n```python\nwinners = remerge.run(corpus, 100, method=remerge.SelectionMethod.npmi, min_count=25)\n```\n\n#### Install\n\nLatest release:\n\n```bash\npip install -U remerge-mwe\n```\n\nFor latest from github:\n\n```bash\npip install git+https://github.com/pmbaumgartner/remerge-mwe.git \n```\n\n#### How it works\n\nThe algorithm operates iteratively in two stages: first, it collects all bigrams of co-occurring `lexemes` in the corpus. A measure is calculated on the set of all bigrams to determine a winner. The two lexemes that comprise the winning bigram are merged into a single lexeme. Instances of that bigram in the corpus are replaced with the merged lexeme. All bigrams are collected again on this new corpus and then the process repeats.\n\nAt initialization, a `lexeme` consists of only a single token, but as the algorithm iterates lexemes become multi-word expressions formed from the winning bigrams. `Lexemes` contain two parts: a `word` which is a tuple of strings, and an `index` which represents the position of that specific token in a MWE. For example, if the winning bigram is `(you, know)`, occurrences of that sequence of lexemes will be replaced with `[(you, know), 0]` and `[(you, know), 1]` in the corpus. When bigrams are counted, only a root lexeme (where the index is 0) can form a bigram, so merged tokens don\'t get double counted. For a more visual explanation of a few iterations assuming specific winners, see the image below.\n\n<img src="explanation.png" alt="An explanation of the remerge algorithm" width="820">\n\n#### Limitations\n\n**No tie-breaking logic** - I found this while testing and comparing to the original reference implementation. If two bigrams are tied for the winning metric, there is no tie-breaking mechanism. Both this implementation and the original implementation simply pick the first bigram from the index with the maximum value. We have slightly different implementation of how the bigram statistics table is created (i.e., the ordering of the index), which makes direct comparisons between implementations difficult.\n\n#### Issues with Original Algorithm\n\n##### Single Bigrams with discontinuities forming from distinct Lexeme positions\n\nOne issue with discontinuities or gaps in the original algorithm is that it did not distinguish the position of a satellite lexeme occurring to the left or right of a bigram with a gap.\n\nTake for example these two example sentences, using `-` to represent an arbitrary token:\n\n```\na b c -\na - c b\n```\n\nAssume in a prior iteration, a winning bigram was `(a, _, c)`, representing the token `a`, a gap of `1`, and then the token `c`. with a gapsize of 1. The past algorithm, run on the above corpus, would count the token `b` twice towards the same n-gram `(a, b, c)`, despite there being two distinct n-grams represented here: `(a, b, c)` and `(a, _, c, b)`.\n\nI think the algorithm is counting on the fact that it would be very rare to encounter this sequence of lexemes in a realistic corpus, where the same word would appear within the gap **and** after the gap. I think this is more of an artifact of this specific example with an unrealistically small vocabulary.\n\n#### References\n\n[^1]: awahl1, MERGE. 2017. Accessed: Jul. 11, 2022. [Online]. Available: https://github.com/awahl1/MERGE\n\n[^2]: A. Wahl and S. Th. Gries, “Multi-word Expressions: A Novel Computational Approach to Their Bottom-Up Statistical Extraction,” in Lexical Collocation Analysis, P. Cantos-Gómez and M. Almela-Sánchez, Eds. Cham: Springer International Publishing, 2018, pp. 85–109. doi: 10.1007/978-3-319-92582-0_5.\n\n[^3]: A. Wahl, “The Distributional Learning of Multi-Word Expressions: A Computational Approach,” p. 190.\n\n[^4]: G. Bouma, “Normalized (Pointwise) Mutual Information in Collocation Extraction,” p. 11.\n\n[^5]: T. Dunning, “Accurate Methods for the Statistics of Surprise and Coincidence,” Computational Linguistics, vol. 19, no. 1, p. 14.\n',
    'author': 'Peter Baumgartner',
    'author_email': '5107405+pmbaumgartner@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
